/*****************************************************************************/
/*                                                                           */
/*  Test Function(s) :      External Definitions and Declarations            */
/*                          External Data - Section 3.1                      */
/*                          Local Complex Types - Section 3.1.2              */
/*  Product and Revision :  Unix 1.2                                         */
/*  Author :                Tina Aleksa                                      */
/*  Source File :           source/extn.3122.c                               */
/*  Revisions :             03/21/85    Sqa     Original UTX-32 1.2 Release  */
/*                                                                           */
/*  Date :                  3/29/85                                          */
/*  Descriptions :          For structures, arrays, fields, and unions       */
/*                          defined and declared ( implicitly by context     */
/*                          or explicitly using the external storage class   */
/*                          specifier) in the same file:                     */
/*                          Verify that the locations are initialized        */
/*                          correctly.                                       */
/*                          The compiler fails if this program cannot be     */
/*                          compiled.                                        */
/*  Test Operation :        This 'c' source file is compiled and executed    */
/*                          under the control of the shell command file      */
/*                           ../shell/extn.31.j                              */
/*                                                                           */
/*  Associated Files :      Any results are reported to the default output   */
/*                          device.                                          */
/*                                                                           */
/*****************************************************************************/
 
#include <stdio.h>
#define TRUE 1;
int fail;
 
char          name[30] = "John Anderson";
unsigned char addr[40] = "6901 West Sunrise Blvd";
unsigned      tel[3] = { 5872900, 5872960 };
short         id_code[5] = { 25, 15, 9, 33, 62 };
int           hr_day[5] = { 8, 8, 10, 12, 8 };
float         hr_rate_wk[5] = { 15.0, 15.0, 12.0, 10.0, 15.0 };
short         hr_wkend[2] = { 10, 12 };
long          sick_wk[52] = { 0, 0, 0, 0, 0, 1, 0, 0, 2 };
long float    hr_rate_wkend[2] = { 20.0, 25.0 };
double        earn_wk[52] = { 500.0, 625.0, 450.0, 875.0 };
unsigned short ins_wk[52] = { 10, 5, 20, 30,10 };
unsigned long vac_hr[52] = { 2, 2, 1, 1, 1, 0, 2 };
 
struct emp_rec {
                 int   emp_no;
                 float salary;
               } emp_data = { 8075, 400.0 };
 
union emp_code {
                int code_a;
                short code_b;
                char code_c;
               } code_emp;
 
struct field   {
                unsigned part1 :  4;
                unsigned part2 : 12;
                unsigned part3 : 16;
               } emp_fields = { 7, 8, 55 };
 
struct tab     {  char *os;
                  struct  { char *project;
                            char *deadline;
                            char *leader;
                            struct { char *engineer;
                                   } person;
                          } proj;
               }  dept = { "UNIX",
                           "CCVS",
                           "01/15/83",
                           "C.NOLL",
                           "S.F.LAW" };
 
/*  The following structure is part of a test for qar8 written against
    UTX 2.0 release.  Problem description:  When an extern declaration
    appears within a statement block for an identifier declared as
    struct symbl *fsym = {0};  the compiler generates an "origin" statement.
*/
 
struct symbl  { char ident[9];
              } ;
struct symbl *fsym = { 0 };
 
main()
{
        FILE *fp;
        fp=fopen("extn_res bloc=y","a");
 
/*  Test 1 -  declare several external variables explicitly using the extern
              storage class specifier.  Array sizes are optional when used
              with the extern declaration.
*/
  {
        extern char          name[];
        extern unsigned char addr[40];
        extern long          sick_wk[];
        extern double        earn_wk[52];
        extern short         id_code[];
        extern struct        emp_data;
        extern union         code_emp;
        extern struct        emp_fields;
        extern struct tab dept;
 
        /* initialize union code_emp                        */
 
        code_emp.code_a = 707;
 
        /* verify that locations are initialized correctly  */
 
        if ( emp_data.emp_no != 8075 || emp_data.salary != 400 )
           { fail = TRUE;
             fprintf(fp,"  ** declaration, local complex types 3.1.2.2-1 class 1 fail  -  exit return code : %d%f\n", emp_data.emp_no,
                      emp_data.salary );
              printf("  ** declaration, local complex types 3.1.2.2-1 class 1 fail  -  exit return code : %d%f\n", emp_data.emp_no,
                      emp_data.salary );
           }
 
        if ( code_emp.code_a != 707 )
           { fail = TRUE;
             fprintf(fp,"  ** declaration, local complex types 3.1.2.2-2 class 1 fail   -  exit return code : %d\n", code_emp.code_a );
             printf("  ** declaration, local complex types 3.1.2.2-2 class 1 fail   -  exit return code : %d\n", code_emp.code_a );
           }
 
        if ( emp_fields.part1 != 7 || emp_fields.part2 != 8 ||
             emp_fields.part3 != 55 )
           { fail = TRUE;
             fprintf(fp,"  ** declaration, local complex types 3.1.2.2-3 class 1 fail  -  exit return code : %d%d%d\n", emp_fields.part1,
                   emp_fields.part2, emp_fields.part3 );
              printf("  ** declaration, local complex types 3.1.2.2-3 class 1 fail  -  exit return code : %d%d%d\n", emp_fields.part1,
                   emp_fields.part2, emp_fields.part3 );
           }
 
        if ( tel[1] != 5872960     || id_code[0] != 25  || hr_day[4] != 8  ||
             hr_rate_wk[1] != 15.0 || hr_wkend[1] != 12 || sick_wk[5] != 1 ||
             hr_rate_wkend[0] != 20.0 || earn_wk[3] != 875.0 ||
             ins_wk[4] != 10         ||  vac_hr[5] != 0 )
           { fail = TRUE;
             fprintf(fp,"  ** declaration, local complex types 3.1.2.2-4 class 1 fail  -  exit return code : %d%d%d%f%d%d%f%f%d%d\n",
                   tel[1], id_code[0], hr_day[4], hr_rate_wk[1], hr_wkend[1],
                   sick_wk[5], hr_rate_wkend[0], earn_wk[3], ins_wk[4],
                   vac_hr[5] );
              printf("  ** declaration, local complex types 3.1.2.2-4 class 1 fail  -  exit return code : %d%d%d%f%d%d%f%f%d%d\n",
                   tel[1], id_code[0], hr_day[4], hr_rate_wk[1], hr_wkend[1],
                   sick_wk[5], hr_rate_wkend[0], earn_wk[3], ins_wk[4],
                   vac_hr[5] );
           }
 
        if ( (compare(name,"John Anderson")) ||
             (compare(addr,"6901 West Sunrise Blvd")) )
           { fail = TRUE;
             fprintf(fp,"  ** external, local complex types 3.1.2.2-5 class 1 fail  -  exit return code : %s %s\n", name, addr );
             printf("  ** external, local complex types 3.1.2.2-5 class 1 fail  -  exit return code : %s %s\n", name, addr );
           }
  }
/*  Test UTX Release 2.0 qar8 */
  {
        extern struct symbl *fsym;
  }
 
  if ( !fail )
    {fprintf(fp,"  .. external, local complex types 3.1.2.2 class 1 pass\n");
     printf("  .. external, local complex types 3.1.2.2 class 1 pass\n");}
  fclose(fp);
}
 
/*****************************************************************************/
/*      COMPARE:  function - compare two given strings                       */
/*                inputs   - two strings, s1 and s2                          */
/*                outputs  - == 0, if s1 is equal to s2                      */
/*                            > 0, if s1 is greater than s2                  */
/*                            < 0, if s1 is less than s2                     */
/*****************************************************************************/
 
compare( s1, s2 )
 
char *s1, *s2;
{
        for( ; *s1 == *s2 ; s1++, s2++ )
           if ( *s1 == '\0' )
              return 0;
        return( *s1 - *s2 );
}
